#!/usr/bin/perl
# -*-cperl-*-
#####################################################################################
## EmpiriST 2015 Shared Task: Official Evaluation Script for Phase II: POS Tagging ##
## https://sites.google.com/site/empirist2015/                                     ##
#####################################################################################

use strict;
use warnings;
use utf8;

use Getopt::Long;
use File::Basename;
use List::Util qw(min);
use File::Basename;

our $GoldDir = undef;
our $DiffDir = undef;
our $ErrorFile = undef;
our $Table = 0;
our $TeamName = undef;
our $Map = 0;
our $ShowMapping = 0;
our $Help = 0;

my $ok = GetOptions(
  "gold|g=s" => \$GoldDir,
  "diff|d=s" => \$DiffDir,
  "errors|e=s" => \$ErrorFile,
  "table|t" => \$Table,
  "name|n=s" => \$TeamName,
  "map|m" => \$Map,
  "show-mapping|sm" => \$ShowMapping,
  "help|h" => \$Help,
);
if ($ShowMapping) {
  $ok = 0 unless @ARGV == 0;
}
else {
  $ok = 0 if $GoldDir && !(@ARGV >= 1);
  $ok = 0 if !$GoldDir && @ARGV != 2;
}

die <<'USAGE' unless $ok && !$Help;
Usage:  perl compare_tagging.perl [OPTIONS] tagged.txt gold.txt
        perl compare_tagging.perl [OPTIONS] system/    gold/
        perl compare_tagging.perl [OPTIONS] -g gold_dir/ tagged1.txt tagged2.txt ...
        perl compare_tagging.perl --show-mapping
Options:
  -g <dir>,       directory with gold standard files (usage 2/3)
    --gold=<dir>
  -d <file/dir>,  directory (usage 2/3) or filename (usage 1) for differences report
    --diff=<file/dir>
  -e <file>,      filename for list of error types
    --errors=<file>
  -t, --table     print TAB-delimited output suitable for copy & paste
  -n <name>          prepend team name <name> in first column (with --table)
    --name=<name>
  -m, --map       map extended tagset to STTS 1.0 for evaluation
  --show-mapping  print STTS 1.0 mapping rules, then exit  
  -h, --help      this help page 

Copyright (C) 2016 by Stefan Evert. Licensed under GPLv3.
USAGE

## define EmpiriST --> STTS 1.0 mapping
our %Mapping = ();
printf "%-10s  %s\n", "gold:", "also accepts these tags:" if $ShowMapping;
while (<DATA>) {
  chomp;
  my ($tag, @stts) = split;
  if ($ShowMapping) {
    printf "%-10s  ", $tag;
    foreach my $s (@stts) {
      printf "%-8s", $s;
    }
    print "\n";
  }
  $Mapping{$tag} = { map {$_ => 1} @stts };
}
exit if $ShowMapping;

if (@ARGV == 2 && -d $ARGV[0] && -d $ARGV[1]) {
  ## usage 2: compare all .txt files in two directories
  my $system_dir = shift @ARGV; # system output files
  $GoldDir = shift @ARGV;       # gold standard
  ## use gold standard directory as reference and complain about missing system output
  opendir(DIR, $GoldDir) or die "Error: Can't open directory $GoldDir/ for reading\n";
  my @filenames = grep {/\.txt$/} readdir DIR;
  closedir DIR;
  @ARGV = (); # set system output files as arguments, then defer to usage 3
  foreach my $file (@filenames) {
    my $system_file = "$system_dir/$file";
    die "Error: System output for $GoldDir/$file not found in $system_dir/\n" unless -f $system_file;
    push @ARGV, $system_file;
  }
}

die "Error: $ErrorFile already exists and is a directory.\n" if $ErrorFile && -d $ErrorFile;  

our %Errors = ();

if ($GoldDir) {
  ## usage 2/3: compare multiple files (with identical names in different directories)
  die "Error: $GoldDir/ must be a directory.\n" unless -d $GoldDir;
  die "Error: $DiffDir/ must be a directory.\n" if $DiffDir && !-d $DiffDir;
  if ($Table) {
    print "team\t" if defined $TeamName;
    printf "%s\t%s\t%s\t%s\n", qw(acc errors tokens file);
  }

  my ($n_correct, $n_tokens) = (0, 0);
  foreach my $TaggedFile (@ARGV) {
    die "Error: input file $TaggedFile does not exist (or isn't a plain file).\n" unless -f $TaggedFile;
    my ($basename, $dir) = fileparse($TaggedFile);
    my $GoldFile = "$GoldDir/$basename";
    die "Error: can't find gold standard file $GoldFile (aborted).\n" unless -f $GoldFile;
    my $DiffFile = ($DiffDir) ? "$DiffDir/$basename" : undef;
    my ($corr, $tokens) = compare_tagging($TaggedFile, $GoldFile, $DiffFile, \%Errors);
    $n_correct += $corr;
    $n_tokens += $tokens;
  }

  if ($Table) {
    printf "%s\t", $TeamName if defined $TeamName;
    printf "%.2f\t%d\t%d\t%s\n", 100 * $n_correct / $n_tokens, $n_tokens - $n_correct, $n_tokens, "TOTAL";
  }
  else {
    print "\nOverall result (micro-averaged):\n";
    printf "    %6.2f%% accuracy:%6d errors /%6d tokens\n", 100 * $n_correct / $n_tokens, $n_tokens - $n_correct, $n_tokens;
  }  
}
else {
  ## usage 1: compare a single pair of files
  die "Error: $DiffDir already exists and is a directory.\n" if $DiffDir && -d $DiffDir;  
  my ($TaggedFile, $GoldFile) = @ARGV;
  if ($Table) {
    print "team\t" if defined $TeamName;
    printf "%s\t%s\t%s\t%s\n", qw(acc errors tokens file);
  }
  compare_tagging($TaggedFile, $GoldFile, $DiffDir, \%Errors);
}

if ($ErrorFile) {
  open(ERRORS, ">:perlio:encoding(utf8)", $ErrorFile) or die "Error: can't write error file $ErrorFile: $!";
  my @pairs = sort {$Errors{$b} <=> $Errors{$a} || $a cmp $b} keys %Errors;
  foreach my $pair (@pairs) {
    print ERRORS "$pair\t$Errors{$pair}\n";
  }
  close ERRORS;
}


sub compare_tagging {
  my ($TaggedFile, $GoldFile, $DiffFile, $ErrorHash) = @_;
  my @TaggedLines = read_lines($TaggedFile);
  my @GoldLines = read_lines($GoldFile);
  if ($DiffFile) {
    open(DIFF, ">:perlio:encoding(utf8)", $DiffFile) or die "Error: can't write diff file $DiffFile: $!";
  }
  my $n_lines = min(@TaggedLines+0, @GoldLines+0);
  my $n_correct = 0;
  foreach my $i (0 .. $n_lines-1) {
    my ($line1, $word1, $pos1) = @{$TaggedLines[$i]};
    my ($line2, $word2, $pos2) = @{$GoldLines[$i]};
    if ($word1 ne $word2) {
      die sprintf "Error: tokenisation mismatch:\n\t%-30s (%s, line #%d)\n\t%-30s (%s, line #%d)\n",
        $word1, $TaggedFile, $line1, $word2, $GoldFile, $line2;
    }
    my $ok = $pos1 eq $pos2;
    $ok = 1 if $Map && $Mapping{$pos2}{$pos1};
    if ($ok) {
      $n_correct++;
      print DIFF "$word1\t$pos1\n" if $DiffFile;
    }
    else {
      $ErrorHash->{"$pos1\t$pos2"}++;
      print DIFF "$word1\t$pos1\t## --> $pos2\n" if $DiffFile;
    }
  }
  close DIFF if $DiffFile;
  if (@TaggedLines > $n_lines) {
    die sprintf "Error: extra lines in file %s (starting at line #%d)\n", $TaggedFile, $TaggedLines[$n_lines][0];
  }
  if (@GoldLines > $n_lines) {
    die sprintf "Error: extra lines in file %s (starting at line #%d)\n", $GoldFile, $GoldLines[$n_lines][0];
  }
  if ($Table) {
    my $basename = basename($TaggedFile, ".txt");
    printf "%s\t", $TeamName if defined $TeamName;
    printf "%.2f\t%d\t%d\t%s\n", 100 * $n_correct / $n_lines, $n_lines - $n_correct, $n_lines, $basename;
  }
  else {
    print "$TaggedFile <-> $GoldFile (gold standard):\n";
    printf "    %6.2f%% accuracy:%6d errors /%6d tokens\n", 100 * $n_correct / $n_lines, $n_lines - $n_correct, $n_lines;
  }
  return $n_correct, $n_lines, 100 * $n_correct / $n_lines;
}


sub read_lines {
  my $filename = shift;
  open(FH, "<:crlf:encoding(utf8)", $filename) or die "Error: can't open input file $filename: $!";
  my @lines = ();
  my @missing = ();
  while (<FH>) {
    chomp;
    s/^\s+//; s/\s+$//;   # trim whitespace
    next if /^<.*>$/;     # skip XML tags (which might include comment markers)
    s/\s*(##|%%).*$//;    # remove comments
    next if /^$/;         # skip blank lines
    next if /^<.*>$/;     # skip XML tags after stripping comments
    my @fields = split /\t/;
    if (@fields == 1) {
      push @missing, $.;
      push @fields, "<missing>";
    }
    elsif (@fields > 2) {
      die "Format error, expected word TAB tag ($filename, line #$.)\n\t".join(" TAB ", @fields)."\n";
    }
    push @lines, [$., @fields];
  }
  close FH;
  if (@missing) {
    my $n = @missing;
    my $pl = ($n > 1) ? "s" : "";
    printf STDERR "Warning: %d missing POS tag%s in file %s\n         line%s %s\n", $n, $pl, $filename, $pl, join(", ", @missing);
  }
  return @lines;
}

## EmpiriST tagset --> STTS 1.0 mapping rules
__DATA__
EMOASC      XY ITJ EMOIMG
EMOIMG      XY ITJ EMOASC
AKW         VVFIN VVIMP VVINF VVIZU VAFIN VAIMP VAINF VMFIN VMINF
HST         XY
ADR         XY NE
URL         XY
EML         XY
VVPPER      VVFIN VVIMP VVINF
VMPPER      VMFIN VMINF
VAPPER      VAFIN VAIMP VAINF
KOUSPPER    KOUS
PPERPPER    PPER
ADVART      ART
PTKIFG      ADV ADJD PTKMA PTKMWL
PTKMA       ADV ADJD PTKIFG PTKMWL
PTKMWL      ADV ADJD PTKIFG PTKMA
DM          KOUS ADV
ONO         ITJ VVFIN VVIMP VVINF
ADV         PTKIFG PTKMA PTKMWL DM
KOUS        DM
PIDAT       PIAT
$(          TRUNC
