#!/usr/bin/perl
# -*-cperl-*-
## Count tokens on separate lines, excluding XML tags and blank lines

use strict;
use warnings;
use utf8;

use Getopt::Long;

our $Delim = undef;
our $Total = 0;
our $Help = 0;

my $ok = GetOptions(
  "delim|d=s" => \$Delim,
  "total|c" => \$Total,
  "help|h" => \$Help,
);

die <<'USAGE' unless $ok && !$Help && @ARGV >= 1;
Usage:  perl line_count.perl [OPTIONS] file.txt ...
Options:
  -d <name>       only count lines within XML region <name>
    --delim=<name>
  -c, --total     also show total number of tokens in all input files
  -h, --help      this help page 
USAGE

my $total_cnt = 0;

foreach my $file (@ARGV) {
  my $cnt = process_file($file);
  printf "%-50.50s  %6d lines\n", $file, $cnt;
  $total_cnt += $cnt;
}

if ($Total) {
  printf "%-50.50s  ------\n", "";
  printf "%-50.50s  %6d\n", "", $total_cnt;
}

## process_file($input, $output);
sub process_file {
  my $input_file = shift;
  my $n = 0;
  my $inside = ($Delim) ? 0 : 1;
  my $n_regions;

  open(IN, "<:encoding(utf8)", $input_file) or die "Can't open input file '$input_file': $!";
  while (<IN>) {
    my $line = $_;
    chomp;
    s/%%.*$//;              # remove annotator comments
    s/^\s+//;               # remove leading/trailing whitespace
    s/\s+$//;
    my $is_xml = (m{^</?\w+.*>$}) ? 1 : 0;
    if ($is_xml && $Delim) {
      if (/^<${Delim}>$/) {
        die "Error: nested <$Delim> regions (${input_file}, line #$.)\n" if $inside;
        $inside = 1;
      }
      elsif (/^<\/${Delim}>$/) {
        die "Error: too many </$Delim> tags (${input_file}, line #$.)\n" if !$inside;
        $inside = 0;
        $n_regions++;
      }
    }
    if ($inside) {
      $n++ unless $is_xml || /^$/;
    }
  }
  close(IN);

  if ($Delim) {
    die "Error: missing </$Delim> tag at end of file (${input_file})\n" if $inside;
    print "Warning: no <$Delim> elements found in input file ${input_file}\n" unless $n_regions > 0;
  }  
  
  return $n;
}
