#!/usr/bin/perl
# -*-cperl-*-
## Normalize text into UTF-8 NFC without non-printing characters
## Optionally, a whitespace-based pre-tokenization can be applied

use strict;
use warnings;
use utf8;

use Unicode::Normalize;
use Getopt::Long;
use Encode::Guess;
use File::Basename;
use Encode;


our $OutDir = undef;
our @Header = ();
our $Split = 0;
our $Verbose = 0;
our $Help = 0;

my $ok = GetOptions(
  "output|o=s" => \$OutDir,
  "header|p=s" => \@Header,
  "split|s" => \$Split,
  "verbose|v" => \$Verbose,
  "help|h" => \$Help,
);
$ok = 0 if $OutDir && !(@ARGV >= 1);
$ok = 0 if !$OutDir && @ARGV != 2;

die <<'USAGE' unless $ok && !$Help;
Usage:  perl normalize_text.perl [OPTIONS] input.txt output.txt
        perl normalize_text.perl [OPTIONS] -o out_dir/ input1.txt input2.txt ...
Options:
  -o <dir>,       directory for output files (usage 2)
    --output=<dir>
  -p '<line>'     prepend header line to file(s) (repeat for multi-line header)
    --header='<line>'
  -s, --split     pre-tokenize text(s) by splitting on whitespace
  -v, --verbose   print names of files being processed
  -h, --help      this help page 
USAGE

if ($OutDir) {
  ## usage 2: process multiple files and store under same name in output directory
  die "Error: $OutDir/ must be a directory.\n" unless -d $OutDir;
  foreach my $InputFile (@ARGV) {
    die "Error: input file $InputFile does not exist (or isn't a plain file).\n" unless -f $InputFile;
    my ($basename, $dir) = fileparse($InputFile);
    my $OutputFile = "$OutDir/$basename";
    process_file($InputFile, $OutputFile, $Split);
  }
}
else {
  ## usage 1: process a single file
  my ($InputFile, $OutputFile) = @ARGV;
  die "Error: input file $InputFile does not exist (or isn't a plain file).\n" unless -f $InputFile;
  process_file($InputFile, $OutputFile, $Split);
}

## process_file($input, $output, $split = 0);
sub process_file {
  my $input_file = shift;
  my $output_file = shift;
  my $split = (@_) ? shift : 0;

  open(IN, "<:perlio", $input_file) or die "Can't open input file '$input_file': $!";
  my $text = join("", <IN>);
  close(IN);
  my $enc = guess_encoding($text);
  die "Error: input file $input_file is not valid UTF-8, please check encoding.\n" unless ref($enc) && $enc->name =~ /^(ascii|utf8)$/;

  open(IN, "<:crlf:encoding(utf8)", $input_file) or die "Can't re-open input file '$input_file': $!";
  open(OUT, ">:perlio:encoding(utf8)", $output_file) or die "Can't open output file '$output_file': $!";

  foreach (@Header) {
    print OUT decode("UTF-8", $_), "\n";
  }
  while (<IN>) {
    chomp;
    if (/^\s*$/) {
      print OUT "\n"; # blank line (may contain whitespace)
    }
    else {
      $_ = NFC($_);           # normalize to NFC
      s/\t/        /g;        # standardize TABs to 8 blanks
      s/\p{C}//g;             # scrub any weird non-printing characters
      s/\p{Diacriticals}//g;  # scrub combining diacritical marks
      next if /^\s*$/; # all content has been removed -> skip to avoid spurious blank line
      if ($split) {
        my $line = $_;
        while ($line =~ m{</?\w+[^<>]*>}) {
          my ($before, $tag) = ($`, $&);
          $line = $';
          foreach (split " ", $before) {
            print OUT "$_\n";
          }
          print OUT "$tag\n";
        }
        foreach (split " ", $line) {
          print OUT "$_\n";
        }
      }
      else {
        print OUT "$_\n";
      }
    }
  }

  close IN;
  close OUT;
  
  print "$input_file --> $output_file\n" if $Verbose;
}