#!/usr/bin/perl
# -*-cperl-*-
############################################################################
## EmpiriST 2015 Shared Task: Validation Script for Phase II: POS Tagging ##
## https://sites.google.com/site/empirist2015/                            ##
############################################################################

use strict;
use warnings;
use utf8;

use Getopt::Long;
use File::Basename;
use List::Util qw(min);

our $NoTags = 0;
our $Comments = 0;
our $Help = 0;
my $ok = GetOptions(
  "no-tags|nt" => \$NoTags,
  "comments|c" => \$Comments,
  "help|h" => \$Help,
);
$ok = 0 unless @ARGV == 2;

die <<'USAGE' unless $ok && !$Help;
Usage:  perl [OPTIONS] validate_tagging.perl system/    tokenized/
        perl [OPTIONS] validate_tagging.perl system.txt tokenized.txt
Options:
  -nt, --no-tags  don't check whether POS tags in system output are valid
  -c, --comments  allow (and skip) annotator comments in tagged file
  -h, --help      this help page 

Copyright (C) 2016 by Stefan Evert. Licensed under GPLv3.
USAGE

our ($System, $Gold) = @ARGV;
our %ValidTag = map {chomp; $_ => 1} <DATA>;

if (-d $Gold && -d $System) {
  ## usage 1: use gold standard directory as reference and check that system output is available for all files
  opendir(DIR, $Gold) or die "Error: Can't open directory $Gold/ for reading\n";
  my @filenames = grep {/\.txt$/} readdir DIR;
  closedir DIR;

  my $Errors = 0;

  foreach my $file (@filenames) {
    my $system_file = "$System/$file";
    my $gold_file = "$Gold/$file";
    print "FILE $file\n";
  
    if (-f $system_file) {
      $Errors++ unless compare_tagging($system_file, $gold_file);
    }
    else {
      print "Error: System output for $gold_file not found in $System/\n";
      $Errors++;
    }
    print "\n\n";
  }
  
  if ($Errors == 0) {
    print "Validation OK -- you're ready to submit\n";
  }
  else {
    printf "ERROR: %d out of %d files had problems (please see above for details)\n", $Errors, scalar @filenames;
  }
}
else {
  ## usage 2: check individual file against gold standard file
  die "Error: either both arguments must be existing directories or both must be plain files\n" unless -f $Gold && -f $System;
  print "Validation OK\n" if compare_tagging($System, $Gold);
}



sub compare_tagging {
  my ($TaggedFile, $GoldFile) = @_;
  my @TaggedLines = read_lines($TaggedFile);
  my @GoldLines = read_lines($GoldFile);
  my $n_lines = min(@TaggedLines+0, @GoldLines+0);
  my $n_errors = 0;
  foreach my $i (0 .. $n_lines-1) {
    my ($line1, $word1, $pos1) = @{$TaggedLines[$i]};
    my ($line2, $word2, $pos2) = @{$GoldLines[$i]};
    if ($word1 ne $word2) {
      printf "Error: tokenisation mismatch:\n\t%-30s (%s, line #%d)\n\t%-30s (%s, line #%d)\n",
        $word1, $TaggedFile, $line1, $word2, $GoldFile, $line2;
      $n_errors++;
    }
    unless ($NoTags) {
      if ($pos1 eq "") {
        printf "Error: missing POS tag in system output (%s, line #%d)\n", $TaggedFile, $line1;
        $n_errors++;
      }
      elsif (!$ValidTag{$pos1}) {
        printf "Error: unknown POS tag '%s' in system output (%s, line #%d)\n", $pos1, $TaggedFile, $line1;
        $n_errors++;
      }
    }
  }
  if (@TaggedLines > $n_lines) {
    printf "Error: extra lines in file %s (starting at line #%d)\n", $TaggedFile, $TaggedLines[$n_lines][0];
    $n_errors += @TaggedLines - $n_lines;
  }
  if (@GoldLines > $n_lines) {
    printf "Error: missing lines in file %s (starting at %s, line #%d)\n", $TaggedFile, $GoldFile, $GoldLines[$n_lines][0];
    $n_errors += @GoldLines - $n_lines;
  }
  printf "WARNING: %d errors found\n", $n_errors if $n_errors > 0;
  return $n_errors == 0;
}


sub read_lines {
  my $filename = shift;
  open(FH, "<:crlf:encoding(utf8)", $filename) or die "Error: can't open input file $filename: $!";
  my @lines = ();
  my @missing = ();
  while (<FH>) {
    chomp;
    s/^\s+//; s/\s+$//;   # trim whitespace
    next if /^<.*>$/;     # skip XML tags (which might include comment markers)
    if ($Comments) {
      ## comments are only allowed in manually annotated files
      s/\s*(##|%%).*$//;    # remove comments
      next if /^<.*>$/;     # skip XML tags after stripping comments
    }
    next if /^$/;         # skip blank lines
    my @fields = split /\t/;
    if (@fields == 1) {
      push @fields, "";  # no POS tag provided
    }
    elsif (@fields > 2) {
      die "Format error, expected word TAB tag ($filename, line #$.)\n\t".join(" TAB ", @fields)."\n";
    }
    push @lines, [$., @fields];
  }
  close FH;
  return @lines;
}


## EmpiriST tagset for validation of system output
__DATA__
ADJA
ADJD
ADV
APPR
APPRART
APPO
APZR
ART
CARD
FM
ITJ
ONO
DM
KOUI
KOUS
KON
KOKOM
NN
NE
PDS
PDAT
PIS
PIAT
PIDAT
PPER
PPOSS
PPOSAT
PRELS
PRELAT
PRF
PWS
PWAT
PWAV
PAV
PTKZU
PTKNEG
PTKVZ
PTKANT
PTKA
PTKIFG
PTKMA
PTKMWL
TRUNC
VVFIN
VVIMP
VVINF
VVIZU
VVPP
VAFIN
VAIMP
VAINF
VAPP
VMFIN
VMINF
VMPP
VVPPER
VMPPER
VAPPER
KOUSPPER
PPERPPER
ADVART
EMOASC
EMOIMG
AKW
HST
ADR
URL
EML
XY
$,
$.
$(
