#!/usr/bin/perl
# -*-cperl-*-
############################################################################
## EmpiriST 2015 Shared Task: Validation Script for Phase I: Tokenization ##
## https://sites.google.com/site/empirist2015/                            ##
############################################################################

use strict;
use warnings;
use utf8;
binmode(STDOUT, ":encoding(utf-8)");

use Unicode::Normalize;
use Algorithm::Diff qw(LCSidx);
use List::Util qw(min sum);
use Getopt::Long;

our $WithBlanks = 0;
our $WithXML = 0;
our $Help = 0;

my $ok = GetOptions(
  "with-blanks|b" => \$WithBlanks,
  "with-xml|x" => \$WithXML,
  "help|h" => \$Help,
);
$ok = 0 unless @ARGV == 2;

die <<'USAGE' unless $ok && !$Help;
Usage:  perl validate_tokenization.perl [OPTIONS] system.txt raw.txt
        perl validate_tokenization.perl [OPTIONS] system/    raw/
Options:
  -b, --with-blanks  include blank lines in comparison (must match)
  -x, --with-xml     include XML tags in comparison (must match)
  -h, --help         this help page 

Copyright (C) 2016 by Stefan Evert. Licensed under GPLv3.
USAGE

our ($File1, $File2) = @ARGV;
my $isdir1 = -d $File1;
my $isdir2 = -d $File2;
die "Error: Either both arguments must be directories, or both must be plain files.\n" unless $isdir1 == $isdir2;

if ($isdir1) {
  opendir(DIR, $File2) or die "Error: Can't open directory $File2/ for reading\n";
  my @filenames = grep {/\.txt$/} readdir DIR;
  closedir DIR;
  my $n_errors = 0;
  foreach my $file (@filenames) {
    print "FILE $file\n";
    if (-f "$File1/$file") {
      $n_errors++ unless compare_files("$File1/$file", "$File2/$file");
    }
    else {
      print "Error: No text file found in $File1/ corresponding to $File2/$file\n";
      $n_errors++;
    }
    print "\n\n";
  }
  if ($n_errors == 0) {
    print "Validation OK -- you're ready to submit\n";
  }
  else {
    printf "ERROR: %d out of %d files had problems (please see above for details)\n", $n_errors, scalar @filenames;
  }
}
else {
  print "Validation OK\n" if compare_files($File1, $File2);
}


## validate tokenized file against raw text input file
##   $ok = compare_files($system_file, $raw_file);
sub compare_files {
  my ($file1, $file2) = @_;
  our ($char1, $bdry1, $lnum1, $lines1, $has_ws1) = read_file($file1);
  our ($char2, $bdry2, $lnum2, $lines2, $has_ws2) = read_file($file2);

  my $n_errors = 0;
  if (@$has_ws1) {
    print "Error: invalid tokens containing whitespace in system output ($file1) on lines\n";
    print "       ", join(", ", @$has_ws1), "\n";
    $n_errors++;
  }

  my $nc1 = @$char1;
  my $nc2 = @$char2;
  my ($i1, $i2) = (0, 0);

  while ($i1 < $nc1 && $i2 < $nc2) {
    my $c1 = $char1->[$i1];
    my $c2 = $char2->[$i2];
    if ($c1 ne $c2) {
      ## character mismatch --> re-synchronize
      printf "Error: character mismatch on line #%d <-> #%d\n", $lnum1->[$i1], $lnum2->[$i2];
      $n_errors++;
      my $span = min(200, $nc1 - $i1, $nc2 - $i2);
      my @chunk1 = @{$char1}[$i1 .. $i1 + $span - 1]; # compare chunks of next 200 chars
      my @chunk2 = @{$char2}[$i2 .. $i2 + $span - 1];
      printf "       System: %-70.70s\n", join("", @chunk1);
      printf "       Input:  %-70.70s\n", join("", @chunk2);
      my ($lcs1, $lcs2) = LCSidx(\@chunk1, \@chunk2);
      if (@$lcs1 == 0) {
        ## synchronization failed, which is acceptable if close to end of text
        print "FATAL ERROR: re-synchronization failed, validation aborted\n" unless $span < 200;
        last;
      }
      my $d1 = $lcs1->[0];
      my $d2 = $lcs2->[0];
      printf "       skipping %d <-> %d characters\n", $d1, $d2;
      $i1 += $d1; # ignore mismatched characters
      $i2 += $d2;
    }
    else {
      $i1++;
      $i2++;
    }
  }

  ## remaining data counts as false positives / negatives
  if ($i1 < $nc1) {
    printf "Error: extra data at end of Text 1, starting on line #%d\n", $lnum1->[$i1];
    $n_errors++;
  }
  if ($i2 < $nc2) {
    printf "Error: extra data at end of Text 2, starting on line #%d\n", $lnum2->[$i2];
    $n_errors++;
  }

  return $n_errors == 0;
}


## read and normalize file, then return as array of chars with token boundary markers
## ($tokens, $bdrys, $line_nums, $lines, $tokens_with_whitespace) = read_file($filename);
sub read_file {
  my $filename = shift;
  my @chars = (); # normalized input text as sequence of chars without token delimiters
  my @bdry = ();  # boolean (0/1) indicating token boundarys (last char of each token)
  my @lnum = ();  # matching list of line numbers in original file
  my @lines = (); # original input lines (normalized)
  my @ws_tokens = (); # input lines with tokens containing whitespace (not allowed)
  open(IN, "<:crlf:encoding(utf8)", $filename) or die "Error: Can't open input file '$filename': $!";
  while (<IN>) {
    chomp;
    $_ = NFC($_);           # normalize to NFC
    s/%%.*$//;              # remove annotator comments
    s/\p{C}//g;             # scrub any weird non-printing characters
## -- seems not to work on ancient Perl versions (5.14.x)
##    s/\p{Diacriticals}//g;  # scrub combining diacritical marks
    s/^\s+//; s/\s+$//;     # remove leading/trailing whitespace
    push @lines, $_;
    next if m{^$} && !$WithBlanks; # skip blank lines unless --with-blanks
    my $is_xml = (m{^</?\w+.*>$}) ? 1 : 0;
    next if $is_xml && !$WithXML; # skip XML tags unless --with-xml
    if (/\s+/) {
      push @ws_tokens, $. unless $is_xml;
      s/\s+//g;               # remove whitespace within tokens      
    }
    $_ = " " if $_ eq "";   # empty tokens (should only happen --with-blanks) recoded as " "
    my $l = length($_);
    push @chars, split //, $_;      # append token to character sequence
    push @bdry, (0) x ($l - 1), 1;  # mark last character in token as boundary
    push @lnum, ($.) x $l;          # original input line for each char
  }
  close IN;
  die "INTERNAL ERROR" unless @chars == @bdry && @chars == @lnum;
  return \@chars, \@bdry, \@lnum, \@lines, \@ws_tokens;
}
